@@ -62,10 +62,9 @@ use List::Compare;
 use List::Util qw(max min any all);
 use Path::Tiny;
 use Test::More;
-use Text::CSV;
+use Text::Diff;
 use Try::Tiny;
 
-use Lintian::Command qw(safe_qx);
 use Lintian::Profile;
 
 use Test::Lintian::ConfigFile qw(read_config);
@@ -73,8 +72,7 @@ use Test::Lintian::Helper qw(rfc822date);
 use Test::Lintian::Hooks
   qw(find_missing_prerequisites run_lintian sed_hook sort_lines calibrate);
 use Test::Lintian::Prepare qw(early_logpath);
-use Test::Lintian::Output::Universal qw(get_tagnames);
-use Test::StagedFileProducer;
+use Test::Lintian::Output::Universal qw(get_tagnames order);
 
 use constant SPACE => q{ };
 use constant EMPTY => q{};
@@ -82,6 +80,11 @@ use constant NEWLINE => qq{\n};
 use constant YES => q{yes};
 use constant NO => q{no};
 
+# turn off the @@-style headers in Text::Diff
+no warnings 'redefine';
+sub Text::Diff::Unified::file_header { return EMPTY; }
+sub Text::Diff::Unified::hunk_header { return EMPTY; }
+
 =head1 FUNCTIONS
 
 =over 4
@@ -110,8 +113,7 @@ sub logged_runner {
             # call runner
             runner($runpath, $betterlogpath)
 
-        }
-        catch {
+        }catch {
             # catch any error
             $error = $_;
         };
@@ -240,172 +242,124 @@ sub runner {
     # set the testing plan
     plan tests => 1;
 
-    my $producer = Test::StagedFileProducer->new(path => $runpath);
-    $producer->exclude(@exclude);
-
     # get lintian subject
     die 'Could not get subject of Lintian examination.'
       unless exists $testcase->{build_product};
     my $subject = "$runpath/$testcase->{build_product}";
 
-    # build subject for lintian examination
-    $producer->add_stage(
-        products => [$subject],
-        minimum_epoch => $threshold,
-        build =>sub {
-            if(exists $testcase->{build_command}) {
-                my $command= "cd $runpath; $testcase->{build_command}";
-                die "$command failed" if system($command);
-            }
-
-            die 'Build was unsuccessful.'
-              unless -f $subject;
-        });
-
     # run lintian
-    my $raw = "$runpath/tags.actual";
-    $producer->add_stage(
-        products => [$raw],
-        minimum_epoch => $lintian_epoch,
-        build =>sub {
-            $ENV{'LINTIAN_COVERAGE'}.= ",-db,./cover_db-$testcase->{testname}"
-              if exists $ENV{'LINTIAN_COVERAGE'};
+    $ENV{'LINTIAN_COVERAGE'}.= ",-db,./cover_db-$testcase->{testname}"
+      if exists $ENV{'LINTIAN_COVERAGE'};
 
-            my $lintian = read_config("$runpath/lintian-command");
-            my $command
-              = "cd $runpath; $ENV{'LINTIAN_FRONTEND'} $lintian->{options} $lintian->{subject}";
-            my ($output, $status) = capture_merged { system($command); };
-            $status = ($status >> 8) & 255;
+    my $lintian = read_config("$runpath/lintian-command");
+    my $command
+      = "cd $runpath; $ENV{'LINTIAN_FRONTEND'} $lintian->{options} $lintian->{subject}";
+    my ($output, $status) = capture_merged { system($command); };
+    $status = ($status >> 8) & 255;
 
-            my @lines = split(NEWLINE, $output);
+    die "$command exited with status $status" . NEWLINE . $output
+      unless $status == 0 || $status == 1;
 
-            if (exists $ENV{LINTIAN_COVERAGE}) {
-                # Devel::Cover causes deep recursion warnings.
-                @lines = grep {
-                    !m{^Deep [ ] recursion [ ] on [ ] subroutine [ ]
-                   "[^"]+" [ ] at [ ] .*B/Deparse.pm [ ] line [ ]
-                   \d+}xsm
-                } @lines;
-            }
+    # filter out some warnings if running under coverage
+    my @lines = split(/\n/, $output);
+    if (exists $ENV{LINTIAN_COVERAGE}) {
+        # Devel::Cover causes deep recursion warnings.
+        @lines = grep {
+            !m{^Deep [ ] recursion [ ] on [ ] subroutine [ ]
+           "[^"]+" [ ] at [ ] .*B/Deparse.pm [ ] line [ ]
+           \d+}xsm
+        } @lines;
+    }
 
-            unless ($status == 0 || $status == 1) {
-                unshift(@lines, "$command exited with status $status");
-                die join(NEWLINE, @lines);
-            }
+    # put output back together
+    $output = EMPTY;
+    $output .= $_ . NEWLINE for @lines;
 
-            # do not forget the final newline, or sorting will fail
-            my $contents
-              = scalar @lines ? join(NEWLINE, @lines) . NEWLINE : EMPTY;
-            path($raw)->spew($contents);
-        });
+    die 'No match strategy defined'
+      unless length $testcase->{match_strategy};
 
-    # run a sed-script if it exists
-    my $parsed = "$runpath/tags.actual.parsed";
-    $producer->add_stage(
-        products => [$parsed],
-        build =>sub {
-            my $script = "$runpath/post_test";
-            if(-f $script) {
-                sed_hook($script, $raw, $parsed);
-            } else {
-                die"Could not copy actual tags $raw to $parsed: $!"
-                  if(system('cp', '-p', $raw, $parsed));
-            }
-        });
-
-    # sort tags
-    my $sorted = "$runpath/tags.actual.parsed.sorted";
-    $producer->add_stage(
-        products => [$sorted],
-        build =>sub {
-            if($testcase->{sort} eq 'yes') {
-                sort_lines($parsed, $sorted);
-            } else {
-                die"Could not copy parsed tags $parsed to $sorted: $!"
-                  if(system('cp', '-p', $parsed, $sorted));
-            }
-        });
-
-    my $specified = "$runpath/tags";
-
-    # create tags if there are none; helps when calibrating new tests
-    path($specified)->touch
-      unless -e $specified;
-
-    # calibrate tags; may write to $sorted
-    my $calibrated = "$runpath/tags.specified.calibrated";
-    $producer->add_stage(
-        products => [$calibrated],
-        build =>sub {
-            my $script = "$runpath/test_calibration";
-            if(-x $script) {
-                calibrate($script, $sorted, $specified, $calibrated);
-            } else {
-                die"Could not copy expected tags $specified to $calibrated: $!"
-                  if(system('cp', '-p', $specified, $calibrated));
-            }
-        });
-
-    # extract expected tags
-    my $expected = "$runpath/tags.specified.calibrated.universal";
-    $producer->add_stage(
-        products => [$expected],
-        build =>sub {
-            my @command = ('tagextract', '-f', 'EWI', $calibrated, $expected);
-            die 'Error executing: ' . join(SPACE, @command) . ": $!"
-              if system(@command);
-        });
-
-    # extract actual tags
-    my $actual = "$runpath/tags.actual.parsed.sorted.universal";
-    $producer->add_stage(
-        products => [$actual],
-        build =>sub {
-            my @command = ('tagextract', '-f', $testcase->{output_format},
-                $sorted, $actual);
-            die 'Error executing: ' . join(SPACE, @command) . ": $!"
-              if system(@command);
-        });
-
-    say EMPTY;
-    $producer->run(verbose => 1);
-
-    my @errors = check_result($testcase, $expected, $actual);
+    my @errors;
+    if ($testcase->{match_strategy} eq 'literal') {
+        @errors = check_literal($testcase, $runpath, $output);
+    } elsif ($testcase->{match_strategy} eq 'tags') {
+        @errors = check_tags($testcase, $runpath, $output);
+    } else {
+        die "Unknown match strategy $testcase->{match_strategy}.";
+    }
 
     my $okay = !(scalar @errors);
 
     if($testcase->{todo} eq 'yes') {
       TODO: {
             local $TODO = 'Test marked as TODO.';
-            ok($okay, 'Lintian tags match for test marked TODO.');
+            ok($okay, 'Lintian passes for test marked TODO.');
         }
         return;
     }
 
-    ok($okay, "Lintian tags match for $testcase->{testname}");
+    ok($okay, "Lintian passes for $testcase->{testname}");
 
     diag $_ . NEWLINE for @errors;
 
-    #    for uncalibrated tests, use tags in spec path, for cut & paste
-    #      $expected = rel2abs("$testcase->{spec_path}/tags")
-    #      unless -e "$runpath/test_calibration";
+    return;
+}
 
-    unless($okay) {
-        my @command = ('tagdiff', $expected, $actual);
-        my ($diff, $status) = capture_merged { system(@command); };
-        $status = ($status >> 8) & 255;
-        die 'Error executing: ' . join(SPACE, @command) . ": $!"
-          if $status;
+sub check_literal {
+    my ($testcase, $runpath, $output) = @_;
 
-        if (length $diff) {
-            path("$runpath/tagdiff")->spew_utf8($diff);
-            diag '--- ' . abs2rel($expected);
-            diag '+++ ' . abs2rel($actual);
-            diag $diff;
-        }
+    # create expected output if it does not exist
+    my $expected = "$runpath/literal";
+    path($expected)->touch
+      unless -e $expected;
+
+    my $raw = "$runpath/literal.actual";
+    path($raw)->spew($output);
+
+    # run a sed-script if it exists
+    my $actual = "$runpath/literal.actual.parsed";
+    my $script = "$runpath/post_test";
+    if(-f $script) {
+        sed_hook($script, $raw, $actual);
+    } else {
+        die"Could not copy actual tags $raw to $actual: $!"
+          if(system('cp', '-p', $raw, $actual));
     }
 
-    return;
+    return check_result($testcase, $runpath, $expected, $actual);
+}
+
+sub check_tags {
+    my ($testcase, $runpath, $output) = @_;
+
+    # create expected tags if there are none; helps when calibrating new tests
+    my $expected = "$runpath/tags";
+    path($expected)->touch
+      unless -e $expected;
+
+    my $raw = "$runpath/tags.actual";
+    path($raw)->spew($output);
+
+    # run a sed-script if it exists
+    my $actual = "$runpath/tags.actual.parsed";
+    my $sedscript = "$runpath/post_test";
+    if(-f $sedscript) {
+        sed_hook($sedscript, $raw, $actual);
+    } else {
+        die"Could not copy actual tags $raw to $actual: $!"
+          if(system('cp', '-p', $raw, $actual));
+    }
+
+    # calibrate tags; may write to $actual
+    my $calibrated = "$runpath/tags.specified.calibrated";
+    my $calscript = "$runpath/test_calibration";
+    if(-x $calscript) {
+        calibrate($calscript, $actual, $expected, $calibrated);
+    } else {
+        die"Could not copy expected tags $expected to $calibrated: $!"
+          if(system('cp', '-p', $expected, $calibrated));
+    }
+
+    return check_result($testcase, $runpath, $calibrated, $actual);
 }
 
 =item check_result(DESC, EXPECTED, ACTUAL)
@@ -417,38 +371,92 @@ tags before calibration. Returns a list of errors, if there are any.
 =cut
 
 sub check_result {
-    my ($testcase, $expectedpath, $actualpath) = @_;
-
-    # fail if tags do not match
-    return 'Tags do not match' if (compare($expectedpath, $actualpath) != 0);
-
-    # no furter checks if the test is not about tags
-    return unless length $testcase->{check};
+    my ($testcase, $runpath, $expectedpath, $actualpath) = @_;
 
     my @errors;
 
-    my $profile = Lintian::Profile->new(undef, [$ENV{LINTIAN_ROOT}]);
+    my @expectedlines = path($expectedpath)->lines;
+    my @actuallines = path($actualpath)->lines;
 
-    # get tags for checks
-    my @related;
-    my @checks = split(SPACE, $testcase->{check});
-    foreach my $check (@checks) {
-        my $checkscript = $profile->get_script($check);
-        die "Unknown Lintian check $check"
-          unless defined $checkscript;
+    push(@expectedlines, NEWLINE)
+      unless @expectedlines;
+    push(@actuallines, NEWLINE)
+      unless @actuallines;
 
-        push(@related, $checkscript->tags);
+    if ($testcase->{match_strategy} eq 'tags') {
+        @expectedlines
+          = reverse sort { order($a) cmp order($b) } @expectedlines;
+        @actuallines
+          = reverse sort { order($a) cmp order($b) } @actuallines;
     }
 
-    @related = sort @related;
+    my $diff = diff(\@expectedlines, \@actuallines, { CONTEXT => 0 });
+    my @difflines = split(/\n/, $diff);
+    chomp @difflines;
 
-    #diag "#Related tag: $_" for @related;
+    # diag "Difflines: $_" for @difflines;
+
+    if(@difflines) {
+
+        if ($testcase->{match_strategy} eq 'literal') {
+            push(@errors, 'Literal output does not match');
+
+        } elsif ($testcase->{match_strategy} eq 'tags') {
+
+            push(@errors, 'Tags do not match');
+
+            @difflines = reverse sort @difflines;
+            my $tagdiff = join(NEWLINE, @difflines) . NEWLINE;
+            path("$runpath/tagdiff")->spew($tagdiff);
+
+        } else {
+            die "Unknown match strategy $testcase->{match_strategy}.";
+        }
+
+        push(@errors, EMPTY);
+
+        push(@errors, '--- ' . abs2rel($expectedpath));
+        push(@errors, '+++ ' . abs2rel($actualpath));
+        push(@errors, @difflines);
+
+        push(@errors, EMPTY);
+    }
+
+    # stop if the test is not about tags
+    return @errors
+      unless $testcase->{match_strategy} eq 'tags';
 
     # get expected tags
     my @expected = sort +get_tagnames($expectedpath);
 
     #diag "=Expected tag: $_" for @expected;
 
+    # look out for tags being tested
+    my @related;
+
+    if (length $testcase->{check}) {
+
+        my $profile = Lintian::Profile->new(undef, [$ENV{LINTIAN_ROOT}]);
+
+        # use tags related to checks declared
+        my @checks = split(SPACE, $testcase->{check});
+        foreach my $check (@checks) {
+            my $checkscript = $profile->get_script($check);
+            die "Unknown Lintian check $check"
+              unless defined $checkscript;
+
+            push(@related, $checkscript->tags);
+        }
+
+        @related = sort @related;
+
+    } else {
+        # otherwise, look for all expected tags
+        @related = @expected;
+    }
+
+    #diag "#Related tag: $_" for @related;
+
     # calculate Test-For and Test-Against; results are sorted
     my $material = List::Compare->new(\@expected, \@related);
     my @test_for = $material->get_intersection;
@@ -462,18 +470,25 @@ sub check_result {
 
     #diag "*Actual tag found: $_" for @actual;
 
-    # find tags not seen; result is sorted
-    my @missing = List::Compare->new(\@test_for, \@actual)->get_Lonly;
-
     # check for blacklisted tags; result is sorted
     my @unexpected
       = List::Compare->new(\@test_against, \@actual)->get_intersection;
 
     # warn about unexpected tags
-    push(@errors, "Tag $_ seen but listed in Test-Against")for @unexpected;
+    if (@unexpected) {
+        push(@errors, 'Unexpected tags:');
+        push(@errors, SPACE . SPACE . $_) for @unexpected;
+        push(@errors, EMPTY);
+    }
+    # find tags not seen; result is sorted
+    my @missing = List::Compare->new(\@test_for, \@actual)->get_Lonly;
 
     # warn about missing tags
-    push(@errors, "Tag $_ listed in Test-For but not seen")for @missing;
+    if (@missing) {
+        push(@errors, 'Missing tags:');
+        push(@errors, SPACE . SPACE . $_) for @missing;
+        push(@errors, EMPTY);
+    }
 
     return @errors;
 }
@@ -483,3 +498,9 @@ sub check_result {
 =cut
 
 1;
+
+# Local Variables:
+# indent-tabs-mode: nil
+# cperl-indent-level: 4
+# End:
+# vim: syntax=perl sw=4 sts=4 sr et
